# Need for Speed - Transformers

Seminar project for the course [(Trans|Lin|Long|...)former: Self-Attention Mechanisms](https://www.cl.uni-heidelberg.de/courses/ws21/sam/)

[TOC]

## 1. (M|N)LP-Mixer :computer:

The work "[MLP-Mixer: An all-MLP Architecture for Vision](https://arxiv.org/pdf/2105.01601.pdf)" introduced an architecture called MLP-Mixer, that relies on multi-layer perceptrons (MLPs), applies it to image classification and offers an alternative to Transformer Networks. There are two different layer types in MLP-Mixer: one MLP-layer that acts on single image segments ("token mixing") and one MLP-layer that acts across those segments ("channel mixing"). For details see the following figure:

<img src="./mlp-mixer.png" width="500px"></img>

The aims of this project are:
* reimplement the network in tensorflow (instead of jax/flax)
* adapt the MLP-Mixer architecture for NLP tasks (➜ NLP-Mixer)
* create a comparable (convolutional) Baseline
* conduct experiments on datasets that are also used by other efficient transformer approaches

## 2. Instructions :pencil:
### Installing the required modules

```
pip install -r requirements.txt
```

### Building arXiv dataset

1. download [cs.AI.rar](https://github.com/LiqunW/Long-document-dataset/raw/master/cs.AI.rar) and [cs.NE.rar](https://github.com/LiqunW/Long-document-dataset/raw/master/cs.AI.rar)
2. extract files and create the following folder structure:
 ```
arxiv/                    
    cs.AI/
        *.txt
    cs.NE/
        *.txt
```
3. set Data.path from the [Dataloader-module](Dataloader.py) to the path where your arxiv folder is (e.g. */home/students/holzinger/tmp/arxiv*):
 ```python
dataloader = Data("arxiv")
dataloader.path = "/home/students/holzinger/tmp/arxiv"
 ```
## 3. Datasets :floppy_disk:

To compare against results from other publications ([Big Bird](https://proceedings.neurips.cc/paper/2020/hash/c8512d142a2d849725f31a9a7a361ab9-Abstract.html), [Linformer](https://arxiv.org/abs/2006.04768) and [Longformer](https://arxiv.org/abs/2004.05150)) the following text classification datasets are used:

| Dataset  | [IMDb](https://ai.stanford.edu/~amaas/data/sentiment/)  | [SST-2](https://nlp.stanford.edu/sentiment/index.html)  | [arXiv](https://github.com/LiqunW/Long-document-dataset)  |
|---|---:|---:|---:|
| #classes  |  2 |  2 | 2 |
| #examples  | 25000  | 67349  | 6007  |
|  avg. words | 233  |  9 | 8096  |
|  max. words |  2470 | 52  |  104698 |


## 4. Experimental Setup :microscope:

For all experiments the same parameters (epochs, learning rate, layer dimensionality, etc.) are used and can be accessed in the [result files](/results/). Only the embedding type (default / pretrained) is changed. As pretrained embeddings *Bert-uncased* from tensorflow-hub is used.

### 4.1 [Baseline](Baseline.py)

Instead of Mlp-Mixer-Blocks, 1D-Convolution-Blocks are used – otherwise the architecture stays the same.

### 4.2 [NLP-Mixer](NLP-Mixer.py)

Adaptation of the MLP-Mixer architecture:  
Instead of image patches we feed the representations (embeddings) of the text into the network. An image *channel* corresponds now to the *embedding dimension* and *patches* to *tokens*. 


## 5. Results :bar_chart:

Accuracy for the text classification datasets:

<table>
    <tr>
        <td></td>
        <td>IMDb</td>
        <td>SST-2</td>
        <td>arXiv</td>
    </tr>
    <tr>
        <td colspan="4" style="text-align: center; vertical-align: middle;">without pretraining</td>
    </tr>
    <tr>
        <td>Baseline</td>
        <td>84.92</td>
        <td>82.45</td>
        <td>90.94</td>
    </tr>
    <tr>
        <td>NLP-Mixer</td>
        <td>85.00</td>
        <td>80.50</td>
        <td>93.43</td>
    </tr>
    <tr>
        <td colspan="4" style="text-align: center; vertical-align: middle;">with pretraining</td>
    </tr>
    <tr>
        <td>Longformer</td>
        <td>95.7</td>
        <td>---</td>
        <td>---</td>
    </tr>
    <tr>
        <td>Linformer</td>
        <td>94.2</td>
        <td>93.4</td>
        <td>---</td>
    </tr>
    <tr>
        <td>BigBird</td>
        <td>95.2</td>
        <td>94.6</td>
        <td>92.3*</td>
    </tr>
    <tr>
        <td>Baseline</td>
        <td>88.40</td>
        <td>91.86</td>
        <td>92.97</td>
    </tr>
    <tr>
        <td>NLP-Mixer</td>
        <td>88.27</td>
        <td>92.89</td>
        <td>95.00</td>
    </tr>
</table>

\*  *all 6 labels*  

## 6. Analysis & Conclusion :bulb:

For the sake of comparison the parameters for both models have been kept the same throughout the experiments. The Baseline started overfitting pretty fast (in contrast to the NLP-Mixer) and therefore only a rather small number of epochs are trained. Also no extensive hyperparameter tuning was performed in order to stay within a limited timeframe. For the imdb dataset the baseline and nlp-mixer-model performed worse in comparison to other transformer architectures by quite a margin, while coming close for sst2. To have a real comparison for the arxiv dataset, a run with all 6 classes would be needed (as classification tasks get naturally more difficult with an increasing number of classes), but was left out for computing time and resource reasons. 

Even though NLP-Mixer performs better overall than the baseline (especially with pretraining), it is surprising how well the simple baseline performs. Therefore it's not entirely clear if the mixing can approximate the self attention between the tokens. Also there is no way to nicely visualize the connection between words as it can be done with the attention mechanism. In addition transformer approaches still perform better in terms of accuracy, but if runtime is a concern, then one of the approaches introduced in this project could be a valid alternative.

## Resources :notebook_with_decorative_cover:

:page_facing_up: Tolstikhin, I. O., Houlsby, N., Kolesnikov, A., Beyer, L., Zhai, X., Unterthiner, T., ... & Dosovitskiy, A. (2021). **Mlp-mixer: An all-mlp architecture for vision**. *Advances in Neural Information Processing Systems*, 34. [[PDF]](https://arxiv.org/pdf/2105.01601.pdf)

:page_facing_up: Johannes Kiesel, Maria Mestre, Rishabh Shukla, Emmanuel Vincent, Payam Adineh, David Corney, Benno Stein, and Martin Potthast. **SemEval-2019 Task 4: Hyperpartisan News Detection**. In *13th International Workshop on Semantic Evaluation (SemEval 2019)*, pages 829-839, June 2019. Association for Computational Linguistics. [[PDF]](https://aclanthology.org/S19-2145.pdf)

:page_facing_up: Maas, A., Daly, R. E., Pham, P. T., Huang, D., Ng, A. Y., & Potts, C. (2011, June). **Learning word vectors for sentiment analysis**. In *Proceedings of the 49th annual meeting of the association for computational linguistics: Human language technologies* (pp. 142-150). [[PDF]](https://aclanthology.org/P11-1015.pdf)

:page_facing_up: Zaheer, M., Guruganesh, G., Dubey, K. A., Ainslie, J., Alberti, C., Ontanon, S., ... & Ahmed, A. (2020). **Big bird: Transformers for longer sequences**. *Advances in Neural Information Processing Systems, 33*, 17283-17297. [[PDF]](https://arxiv.org/pdf/2007.14062.pdf)

:page_facing_up: Wang, S., Li, B. Z., Khabsa, M., Fang, H., & Ma, H. (2020). **Linformer: Self-attention with linear complexity**. *arXiv preprint arXiv:2006.04768.* [[PDF]](https://arxiv.org/pdf/2006.04768.pdf)

:page_facing_up: Beltagy, I., Peters, M. E., & Cohan, A. (2020). **Longformer: The long-document transformer**. *arXiv preprint arXiv:2004.05150.* [[PDF]](https://arxiv.org/pdf/2004.05150.pdf)

## LICENSE

This project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details